import NDK, { NDKEvent, type NDKEventId, NDKKind, type NDKUser } from '@nostr-dev-kit/ndk';
import { isValidNip05 } from '$lib/utils.svelte';
import { TokenInfoWithMailSubscriptionDuration } from '@arx/utils';

function parseTextPlainContent(mimeMessage: string) {
	const boundary = mimeMessage.split('\n')[0];

	const boundaries = mimeMessage.split(boundary + '\n');
	boundaries.shift();

	const bodies = [];

	for (const body of boundaries) {
		const headers: Record<string, string> = {};
		const lines = body.split(/\r?\n/);
		let isLookingForHeaders = true;
		let content = '';
		for (const line of lines) {
			if (line.trim() === '') {
				isLookingForHeaders = false;
				content += line + '\n';
			} else {
				if (!isLookingForHeaders) {
					content += line + '\n';
					continue;
				}
				const [key, value] = line.split(': ');
				headers[key.toLowerCase()] = value;
			}
		}
		bodies.push({ headers, content: content.trim() });
	}

	if (bodies.length === 0) return mimeMessage;

	if (bodies[0].headers['content-type']?.toLowerCase().startsWith('multipart/alternative'))
		return parseTextPlainContent(bodies[0].content);

	let chosenBody = bodies.find((body) =>
		body.headers['content-type']?.toLowerCase()?.startsWith('text/plain')
	);
	if (!chosenBody) chosenBody = bodies[0];

	const encoding = chosenBody.headers['content-transfer-encoding'];
	if (!encoding) return chosenBody.content; // assume plain text

	switch (encoding.toLowerCase()) {
		case '7bit':
			return chosenBody.content;
		case 'base64':
			const raw = atob(chosenBody.content.replaceAll('\n', ''));
			const bytes = new Uint8Array(raw.length);
			for (let i = 0; i < raw.length; i++) bytes[i] = raw.charCodeAt(i);
			const decoder = new TextDecoder('utf8');
			return decoder.decode(bytes);
		case 'quoted-printable':
			return chosenBody.content
				.replace(/=\r\n/g, '')
				.replace(/=([0-9A-F]{2})/g, (_, p1) => String.fromCharCode(parseInt(p1, 16)));
		default:
			return `Cannot decode ${encoding} encoded message. This is a bug, please report it.`;
	}
}

export class Letter {
	id: NDKEventId;
	subject: string;
	from: NDKUser;
	content: string;
	date: Date;
	recipients: Set<NDKUser>;
	stamp?: TokenInfoWithMailSubscriptionDuration;
	emailAddress?: string;

	private constructor() {
		this.id = '';
		this.subject = '';
		this.from = {} as NDKUser;
		this.content = '';
		this.date = new Date();
		this.recipients = new Set<NDKUser>();
	}

	static async fromDecryptedMessage(
		decryptedMessage: NDKEvent,
		encryptedMessage: NDKEvent,
		ndk: NDK
	): Promise<Letter> {
		if (decryptedMessage.kind !== NDKKind.Article) throw new Error('Not a letter');

		let rawContent = decryptedMessage.content;
		let content = '';
		let email = decryptedMessage.tags.find((t) => t[0] === 'email:from')?.[1];
		if (rawContent.startsWith('--') && email) content = parseTextPlainContent(rawContent);
		else content = rawContent;

		const letter = new Letter();
		letter.id = encryptedMessage.id;
		letter.subject = decryptedMessage.tags.find((t) => t[0] === 'subject')?.[1] ?? 'No subject';
		letter.from = decryptedMessage.author;
		letter.content = content;
		letter.date = new Date(decryptedMessage.created_at! * 1000);
		letter.emailAddress = email;
		let stampToken = decryptedMessage.tags.find((t) => t[0] === 'stamp')?.[1];
		if (stampToken) {
			try {
				stampToken = await ndk.signer!.nip44Decrypt(letter.from, stampToken);
				letter.stamp = new TokenInfoWithMailSubscriptionDuration(stampToken);
			} catch (error) {
				console.error('Error decrypting stamp token:', error);
			}
		}

		for (const tag of decryptedMessage.tags) {
			if (tag[0] === 'p') {
				try {
					if (isValidNip05(tag[1])) {
						const parsed = await ndk.getUserFromNip05(tag[1]);
						if (parsed) letter.recipients.add(parsed);
					} else if (tag[1].startsWith('npub')) {
						letter.recipients.add(ndk.getUser({ npub: tag[1] }));
					} else {
						letter.recipients.add(ndk.getUser({ pubkey: tag[1] }));
					}
				} catch (error) {
					console.error('Error processing recipient:', error);
				}
			}
		}

		return letter;
	}
}
