import type { FolderLabel } from '$lib/folderLabel';
import { type NDKEvent, type NDKEventId } from '@nostr-dev-kit/ndk';
import { decryptSealedMessageIntoReadableType, waitForNDK } from '$lib/utils.svelte';
import type { Letter } from '$lib/letter';

export enum partialLetterStatus {
	NON_LETTER,
	CONTENT_LOADED,
	DECRYPTED
}

export type PartialLetter =
	| { id: NDKEventId; status: partialLetterStatus.NON_LETTER }
	| {
			id: NDKEventId;
			status: partialLetterStatus.CONTENT_LOADED;
			event: NDKEvent;
	  }
	| {
			id: NDKEventId;
			status: partialLetterStatus.DECRYPTED;
			event: Letter;
	  };

export async function* getLettersInFolder(
	folder: FolderLabel,
	ignoreIDs: NDKEventId[] = []
): AsyncGenerator<PartialLetter> {
	const ndk = await waitForNDK();
	const filter =
		folder.id != 'inbox' && folder.id != 'sent'
			? {
					ids: folder.letters
				}
			: {
					kinds: [1059],
					'#p': [ndk.activeUser!.pubkey]
				};
	let letterEvents = [...(await ndk.fetchEvents(filter))].filter((e) => !ignoreIDs.includes(e.id));
	return yield* parseLetters(new Set(letterEvents));
}

async function* parseLetters(letterEvents: Set<NDKEvent>): AsyncGenerator<PartialLetter> {
	for (let event of letterEvents)
		yield {
			id: event.id,
			status: partialLetterStatus.CONTENT_LOADED,
			event
		};
	for (let event of letterEvents) {
		const letter = await decryptSealedMessageIntoReadableType(event);
		if (!letter)
			yield {
				id: event.id,
				status: partialLetterStatus.NON_LETTER
			};
		else
			yield {
				id: event.id,
				status: partialLetterStatus.DECRYPTED,
				event: letter
			};
	}
}

class LetterNode {
	letter: PartialLetter;
	left: LetterNode | null;
	right: LetterNode | null;

	constructor(letter: PartialLetter) {
		this.letter = letter;
		this.left = null;
		this.right = null;
	}
}

class LetterTree {
	root: LetterNode | null;

	constructor() {
		this.root = null;
	}

	insert(letter: PartialLetter) {
		if (this.root) {
			const existingLetter = this.findLetter(this.root, letter.id);
			if (existingLetter) this.remove(letter.id);
		}
		const newNode = new LetterNode(letter);
		if (!this.root) this.root = newNode;
		else this.doInsertLetter(this.root, newNode);
	}

	remove(id: NDKEventId) {
		this.root = this.removeNode(this.root, id);
	}

	removeNode(node: LetterNode | null, id: NDKEventId): LetterNode | null {
		if (!node) return null;

		if (node.letter.id === id) {
			if (!node.left && !node.right) return null;
			else if (!node.left) return node.right;
			else if (!node.right) return node.left;
			else {
				const minNode = this.findMinNode(node.right);
				node.letter = minNode.letter;
				node.right = this.removeNode(node.right, minNode.letter.id);
			}
		} else {
			node.left = this.removeNode(node.left, id);
			node.right = this.removeNode(node.right, id);
		}

		return node;
	}

	findMinNode(node: LetterNode): LetterNode {
		while (node.left) node = node.left;
		return node;
	}

	toSorted(): PartialLetter[] {
		const result: PartialLetter[] = [];
		this.inOrderTraversalNode(this.root, result);
		return result;
	}

	private findLetter(node: LetterNode | null, id: NDKEventId): LetterNode | null {
		if (!node) {
			return null;
		}
		if (node.letter.id === id) {
			return node;
		}
		const leftResult = this.findLetter(node.left, id);
		if (leftResult) {
			return leftResult;
		}
		return this.findLetter(node.right, id);
	}

	private doInsertLetter(letter: LetterNode, newLetter: LetterNode) {
		if (
			letter.letter.status === partialLetterStatus.NON_LETTER ||
			newLetter.letter.status === partialLetterStatus.NON_LETTER
		)
			return;

		let dateA, dateB;

		if (newLetter.letter.status === partialLetterStatus.CONTENT_LOADED)
			dateA = newLetter.letter.event.created_at!;
		else dateA = newLetter.letter.event.date.getTime() / 1000; // Convert Date to Unix timestamp

		if (letter.letter.status === partialLetterStatus.CONTENT_LOADED)
			dateB = letter.letter.event.created_at!;
		else dateB = letter.letter.event.date.getTime() / 1000;

		if (dateA > dateB) {
			if (!letter.left) letter.left = newLetter;
			else this.doInsertLetter(letter.left, newLetter);
		} else {
			if (!letter.right) letter.right = newLetter;
			else this.doInsertLetter(letter.right, newLetter);
		}
	}

	private inOrderTraversalNode(node: LetterNode | null, result: PartialLetter[]) {
		if (node) {
			this.inOrderTraversalNode(node.left, result);
			result.push(node.letter);
			this.inOrderTraversalNode(node.right, result);
		}
	}
}

export async function* sortedLetters(
	folder: FolderLabel,
	ignoreIDs: NDKEventId[] = []
): AsyncGenerator<PartialLetter[]> {
	const ndk = await waitForNDK();
	const lt = new LetterTree();

	for await (const letter of getLettersInFolder(folder, ignoreIDs)) {
		if (letter.status === partialLetterStatus.NON_LETTER) lt.remove(letter.id);
		else {
			const stat = [
				folder.id == 'sent',
				letter.status == partialLetterStatus.DECRYPTED &&
					letter.event.from.pubkey == ndk.activeUser!.pubkey
			];
			if ((stat[0] && !stat[1]) || (!stat[0] && stat[1])) lt.remove(letter.id);
			else lt.insert(letter);
		}

		yield lt.toSorted();
	}
}
